# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import annotations

from datetime import datetime
import logging
import os
import pathlib
import shutil
import typing
import uuid

import httpx

from . import core
from .. import errors, model, utils
from .._typing_compat import override


class ResourceCacheRepository(core.RepositoryContainer):
    """
    A cache for resources based on etags. It stores temporarily
    resources with an assigned etag on the local disk.
    When fallback_to_cache is enabled (default), then a cached resource
    will be returned if the source repository is unavailable. Used in
    the right context, this can be used for example to maintain a functional
    repository for previously seen responses when PyPI.org is down.
    """
    def __init__(
        self,
        source: core.SimpleRepository,
        cache_path: pathlib.Path,
        http_client: typing.Optional[httpx.AsyncClient] = None,
        logger: logging.Logger = logging.getLogger(__name__),
        fallback_to_cache: bool = True,
    ) -> None:
        super().__init__(source)
        self._cache_path = cache_path.resolve()
        self._tmp_path = self._cache_path / ".incomplete"
        self._tmp_path.mkdir(parents=True, exist_ok=True)
        self._http_client = http_client or httpx.AsyncClient()
        self._logger = logger
        self._fallback_to_cache = fallback_to_cache

    @override
    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.Resource:
        """
        Get a resource from the cache. If it is not present in the
        cache, or its etag has expired, retrieve it from the source repository.
        If it's a remote resource, download the resource and cache it.
        """
        request_context = request_context or model.RequestContext()

        project_dir = (self._cache_path / project_name).resolve()
        resource_path = (project_dir / resource_name).resolve()
        resource_info_path = resource_path.with_suffix(resource_path.suffix + ".info")

        # Ensures that the requested resource is contained
        # in the cache directory to avoid path traversal.
        if (
            not utils.is_relative_to(resource_path, self._cache_path) or
            not utils.is_relative_to(project_dir, self._cache_path)
        ):
            raise ValueError(f"{resource_path} is not contained in {self._cache_path}")

        project_dir.mkdir(exist_ok=True)

        # Require the resource upstream, if available use the cached etag.
        cache_etag = resource_info_path.read_text() if resource_info_path.is_file() else None
        if cache_etag:
            context: typing.Mapping[str, typing.Any] = {
                **request_context.context,
                "etag": cache_etag,
            }
        else:
            context = request_context.context
        new_request_context = model.RequestContext(
            context=context,
        )
        try:
            resource = await super().get_resource(
                project_name,
                resource_name,
                request_context=new_request_context,
            )
        except model.NotModified:
            # The upstream repository serves the same content that has been cached.
            # If the request also provides the same etag, raise NotModified,
            # otherwise return the locally cached resource.
            if cache_etag == request_context.context.get("etag"):
                raise
            return self._cached_resource(
                resource_path=resource_path,
                resource_info_path=resource_info_path,
                context=model.Context(etag=cache_etag) if cache_etag else model.Context(),
            )
        except errors.SourceRepositoryUnavailable:
            if not self._fallback_to_cache:
                raise
            # If the source repository behaves incorrectly and we
            # have a cached artifact, return it to the client.
            self._logger.error(
                f"Upstream unavailable, served cached {project_name}:{resource_name}",
            )
            if cache_etag:
                return self._cached_resource(
                    resource_path=resource_path,
                    resource_info_path=resource_info_path,
                    context=model.Context(etag=cache_etag),
                )
            else:
                raise

        upstream_etag = resource.context.get("etag")
        if not upstream_etag or resource.to_cache is False:
            # Only cache a resource if has an etag and to_cache is set to True.
            return resource

        if upstream_etag != cache_etag:
            await self._store_resource(
                resource=resource,
                upstream_etag=upstream_etag,
                resource_path=resource_path,
                resource_info_path=resource_info_path,
            )

        return self._cached_resource(
            resource_path=resource_path,
            resource_info_path=resource_info_path,
            context=resource.context,
        )

    async def _store_resource(
        self,
        resource: model.Resource,
        upstream_etag: str,
        resource_path: pathlib.Path,
        resource_info_path: pathlib.Path,
    ) -> None:
        if isinstance(resource, model.HttpResource):
            # The upstream resource changed or no cached version is available.
            # Fetch the resource and cache it and its etag.
            timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
            dest_file = self._tmp_path / f"{timestamp}_{uuid.uuid4().hex}"
            await utils.download_file(
                download_url=resource.url,
                dest_file=dest_file,
                http_client=self._http_client,
            )
            dest_file.rename(resource_path)
        elif isinstance(resource, model.TextResource):
            resource_path.write_text(resource.text)
        elif isinstance(resource, model.LocalResource):
            shutil.copy(resource.path, resource_path)
        else:
            raise ValueError(f"Unknown resource type: {type(resource)}.")
        resource_info_path.write_text(upstream_etag)

    def _cached_resource(
        self,
        resource_path: pathlib.Path,
        resource_info_path: pathlib.Path,
        context: model.Context,
    ) -> model.LocalResource:
        self._update_last_access(resource_info_path)
        local_resource = model.LocalResource(path=resource_path)
        local_resource.context.update(context)
        return local_resource

    def _update_last_access(
        self,
        resource_info_path: pathlib.Path,
    ) -> None:
        """
        Store the last access as the access and modified times of the file.
        That information will be used to delete unused files in the cache.
        """
        now = datetime.now().timestamp()

        os.utime(resource_info_path, (now, now))



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import annotations

import dataclasses
from datetime import timedelta
import typing
from urllib.parse import urlsplit, urlunsplit

import httpx

from . import core
from .. import errors, model, parser, utils
from .._typing_compat import override


def _url_path_append(url: str, append_with: str) -> str:
    """
    Append to the path part of the URL, taking care of trailing slashes on
    the root_url. `append_with` may contain sub-paths, and may end with a
    slash if desired.

    """
    PATH_IDX = 2

    split_url = list(urlsplit(url))

    if not split_url[PATH_IDX].endswith('/'):
        # Add a trailing slash to the path.
        split_url[PATH_IDX] += '/'

    if append_with.startswith('/'):
        append_with = append_with[1:]

    split_url[PATH_IDX] += append_with
    new_url = urlunsplit(split_url)
    return new_url


class HttpRepository(core.SimpleRepository):
    """Proxy of a remote simple repository"""

    def __init__(
        self,
        url: str,
        http_client: typing.Optional[httpx.AsyncClient] = None,
        connection_timeout: timedelta = timedelta(seconds=15),
    ) -> None:
        self._source_url = url
        self._http_client = http_client or httpx.AsyncClient()
        self.downstream_content_types = ", ".join([
            "application/vnd.pypi.simple.v1+json",
            "application/vnd.pypi.simple.v1+html;q=0.2",
            "text/html;q=0.01",
        ])
        self._connection_timeout = connection_timeout

    async def _fetch_simple_page(
        self,
        page_url: str,
    ) -> typing.Tuple[str, str]:
        """Retrieves a simple page from the given url.
        Returns the body and the content type received.
        """
        headers = {"Accept": self.downstream_content_types}
        response = await self._http_client.get(
            url=page_url,
            headers=headers,
            timeout=self._connection_timeout.total_seconds(),
        )
        response.raise_for_status()
        body = response.text
        content_type: str = response.headers.get("content-type", "")
        return body, content_type

    @override
    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.ProjectDetail:
        page_url = _url_path_append(self._source_url, f'{project_name}/')
        try:
            body, content_type = await self._fetch_simple_page(page_url)
        except httpx.HTTPError as e:
            # If the status_code is 404, the source repository is working correctly, but
            # the requested resource is not available. Any other 4xx or 5xx error code is
            # treated as a source repository misbehaviour
            if isinstance(e, httpx.HTTPStatusError) and e.response.status_code == 404:
                raise errors.PackageNotFoundError(
                    package_name=project_name,
                ) from e
            # This code path also includes connection failures or timeouts.
            raise errors.SourceRepositoryUnavailable() from e

        if (
            "application/vnd.pypi.simple.v1+html" in content_type or
            "text/html" in content_type or not content_type
        ):
            project_page = parser.parse_html_project_page(body, project_name)
        elif "application/vnd.pypi.simple.v1+json" in content_type:
            project_page = parser.parse_json_project_page(body)
        else:
            raise errors.UnsupportedSerialization(content_type)

        # Make the URLs in the project page absolute, such that they can be
        # resolved upstream without knowing the original source URLs.
        files = tuple(
            dataclasses.replace(file, url=utils.url_absolutizer(file.url, page_url))
            for file in project_page.files
        )
        project_page = dataclasses.replace(project_page, files=files)
        return project_page

    @override
    async def get_project_list(
        self,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.ProjectList:
        try:
            body, content_type = await self._fetch_simple_page(self._source_url)
        except httpx.HTTPError as e:
            raise errors.SourceRepositoryUnavailable() from e

        if (
            "application/vnd.pypi.simple.v1+html" in content_type or
            "text/html" in content_type or not content_type
        ):
            return parser.parse_html_project_list(body)
        elif "application/vnd.pypi.simple.v1+json" in content_type:
            return parser.parse_json_project_list(body)

        raise errors.UnsupportedSerialization(content_type)

    @override
    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.Resource:
        request_context = request_context or model.RequestContext()
        try:
            project_page = await self.get_project_page(
                project_name,
                request_context=request_context,
            )
        except errors.PackageNotFoundError:
            raise

        resource: typing.Optional[model.HttpResource] = None
        if resource_name.endswith(".metadata"):
            resource = await self.get_metadata(project_page, resource_name)
        else:
            for file in project_page.files:
                if resource_name == file.filename:
                    resource = model.HttpResource(url=file.url)
                    break

        if not resource:
            raise errors.ResourceUnavailable(resource_name)

        try:
            resp = await self._http_client.head(
                url=resource.url,
                timeout=self._connection_timeout.total_seconds(),
            )
            resp.raise_for_status()
        except httpx.HTTPError as e:
            raise errors.SourceRepositoryUnavailable() from e
        etag = resp.headers.get("ETag")
        if etag:
            if etag == request_context.context.get("etag"):
                # If the etag served from the source repository
                # matches the one in the request raise NotModified
                raise model.NotModified()
            resource.context["etag"] = etag

        return resource

    async def get_metadata(
        self,
        project_page: model.ProjectDetail,
        resource_name: str,
    ) -> model.HttpResource:
        distribution_name = utils.remove_suffix(resource_name, ".metadata")
        for file in project_page.files:
            if distribution_name == file.filename and file.dist_info_metadata:
                return model.HttpResource(url=file.url + ".metadata")
        raise errors.ResourceUnavailable(resource_name)



import typing

from ._frozen_json import FrozenJSONType, JSONMapping
from ._typing_compat import override


class PrivateMetadataMapping(JSONMapping):
    @override
    def _finalize_data(
        self,
        data: typing.Mapping[str, FrozenJSONType],
    ) -> typing.Mapping[str, FrozenJSONType]:
        data = super()._finalize_data(data)
        invalid_keys = ", ".join(
            sorted(repr(key) for key in data if not key.startswith("_")),
        )
        if invalid_keys:
            raise ValueError(
                f"The following keys are invalid for private metadata: {invalid_keys}",
            )
        return data



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import annotations

import typing

import pytest

from ... import errors, model
from ...components.deny_list import DenyListRepository

if typing.TYPE_CHECKING:
    from .fake_repository import FakeRepository


@pytest.fixture
def repository(source_repository: FakeRepository) -> DenyListRepository:
    return DenyListRepository(
        source=source_repository,
        deny_list=("project3"),
    )


@pytest.mark.asyncio
async def test__get_project_list(repository: DenyListRepository) -> None:
    project_list = await repository.get_project_list()
    assert project_list.projects == frozenset([
        model.ProjectListElement("project1"),
        model.ProjectListElement("project2"),
    ])


@pytest.mark.asyncio
async def test__get_project_page__not_deny_listed(
    repository: DenyListRepository,
    source_repository: FakeRepository,
) -> None:
    project_page = await repository.get_project_page("project1")
    assert project_page == source_repository.project_pages["project1"]


@pytest.mark.asyncio
async def test__get_project_page__deny_listed(repository: DenyListRepository) -> None:
    with pytest.raises(errors.PackageNotFoundError):
        await repository.get_project_page("project3")


@pytest.mark.asyncio
async def test__get_resource__not_deny_listed(repository: DenyListRepository) -> None:
    resource = await repository.get_resource("project1", "project1-1.0.tar.gz")
    assert isinstance(resource, model.HttpResource)
    assert resource.url == "content1"


@pytest.mark.asyncio
async def test__get_resource__deny_listed(repository: DenyListRepository) -> None:
    with pytest.raises(errors.ResourceUnavailable):
        await repository.get_resource("project3", "project3-1.0.tar.gz")



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import annotations

import dataclasses
import typing

from . import core
from .. import errors, model
from .._typing_compat import override


class AllowListRepository(core.RepositoryContainer):
    def __init__(self, source: core.SimpleRepository, allow_list: typing.Tuple[str, ...]) -> None:
        super().__init__(source)
        self._allow_list = allow_list

    @override
    async def get_project_list(
        self,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.ProjectList:
        project_list = await super().get_project_list(request_context=request_context)
        projects = frozenset(
            elem for elem in project_list.projects if elem.normalized_name in self._allow_list
        )
        return dataclasses.replace(project_list, projects=projects)

    @override
    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.ProjectDetail:
        if project_name not in self._allow_list:
            raise errors.PackageNotFoundError(project_name)
        return await super().get_project_page(project_name, request_context=request_context)

    @override
    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.Resource:
        if project_name not in self._allow_list:
            raise errors.ResourceUnavailable(resource_name)
        return await super().get_resource(
            project_name, resource_name, request_context=request_context,
        )



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import sys
import typing

if typing.TYPE_CHECKING:
    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

if sys.version_info >= (3, 8):
    from typing import Protocol, TypedDict
else:
    from typing_extensions import Protocol, TypedDict

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

__all__ = [
    'override',
    'Self',
    'TypedDict',
    'TypeAlias',
    'Protocol',
]



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import annotations

import hashlib
import json
import pathlib
import sys
import typing
from urllib.parse import urljoin, urlparse

import httpx

from . import errors


def url_absolutizer(url: str, url_base: str) -> str:
    """Converts a relative url into an absolute one"""
    if not urlparse(url).scheme:
        return urljoin(url_base, url)
    return url


def load_config_json(json_file: pathlib.Path) -> typing.Dict[typing.Any, typing.Any]:
    try:
        json_config = json.loads(json_file.read_text())
    except json.JSONDecodeError as e:
        raise errors.InvalidConfigurationError("Invalid json file") from e
    except FileNotFoundError as e:
        raise errors.InvalidConfigurationError("Configuration file not found") from e
    if not isinstance(json_config, dict):
        raise errors.InvalidConfigurationError(
            f"Invalid configuration file. {str(json_file)}"
            " must contain a dictionary.",
        )
    return json_config


async def download_file(
    download_url: str,
    dest_file: pathlib.Path,
    http_client: httpx.AsyncClient,
    chunk_size: int = 1024 * 64,
) -> None:
    with dest_file.open('wb') as file:
        async with http_client.stream("GET", download_url) as data:
            async for chunk in data.aiter_bytes(chunk_size):
                file.write(chunk)


def remove_prefix(source: str, prefix: str) -> str:
    """Compatibility for pre-3.9 implementations that do not have str.removeprefix"""
    if sys.version_info >= (3, 9):
        return source.removeprefix(prefix)
    if source.startswith(prefix):
        return source[len(prefix):]


def remove_suffix(source: str, suffix: str) -> str:
    """Compatibility for pre-3.9 implementations that do not have str.removesuffix"""
    if sys.version_info >= (3, 9):
        return source.removesuffix(suffix)
    if source.endswith(suffix):
        return source[:-len(suffix)]


def is_relative_to(target: pathlib.Path, match: typing.Union[str, pathlib.Path]) -> bool:
    """Compatibility for pre-3.9 implementations that do not have Path.is_relative_to"""
    if sys.version_info >= (3, 9):
        return target.is_relative_to(match)
    try:
        _ = target.relative_to(match)
        return True
    except ValueError:
        return False


def hash_md5(data: bytes) -> str:
    """Compatibility for pre-3.9 implementations that do not allow md5() call with arguments"""
    if sys.version_info >= (3, 9):
        return hashlib.md5(data, usedforsecurity=False).hexdigest()
    h = hashlib.md5()
    h.update(data)
    return h.hexdigest()



from __future__ import annotations

from collections.abc import Mapping
import sys
import typing

from ._typing_compat import TypeAlias

if sys.version_info >= (3, 10):
    from types import NoneType
else:
    NoneType = type(None)

if typing.TYPE_CHECKING:
    from ._typing_compat import Self

FrozenJSONType: TypeAlias = typing.Union[
    str,
    int,
    float,
    bool,
    None,
    typing.Tuple["FrozenJSONType", ...],
    "JSONMapping",
]


class JSONMapping(typing.Mapping[str, FrozenJSONType]):
    """
    A frozen JSON mapping.

    Note: JSONMapping is hashable (possible since it is frozen).

    Note:

        The JSONMapping type supports the dict union interface, meaning that you can update
        specific keys in the mapping (into a new instance) with:

            new_mapping = mapping | {'some-new-value': 's'}

        You can apply comprehensions indirectly:

            new_mapping = JSONMapping(
                {key: value for key, value in mapping.items() if key not in ['some-new-value']}
            )

    Note that this class also supports `__contains__`, `keys`, `items`, `values`, `get` and
    `__eq__`.

    """

    def __init__(
        self,
        items: typing.Union[
            typing.Tuple[
                typing.Tuple[str, FrozenJSONType],
                ...,
            ],
            typing.Mapping[str, FrozenJSONType],
            None,
        ] = None,
    ):
        self._data: typing.Mapping[str, FrozenJSONType] = self._finalize_data(
            dict(items or ()),
        )

    def _finalize_data(
        self,
        data: typing.Mapping[str, FrozenJSONType],
    ) -> typing.Mapping[str, FrozenJSONType]:
        """
        A hook to allow subclasses to refine the internal data. This could be used for validation or
        default generation, for example.
        """
        return data

    @classmethod
    def from_any_mapping(
        cls,
        data: typing.Mapping[str, typing.Any],
        sub_mapping_cls: typing.Optional[typing.Type[JSONMapping]] = None,
    ) -> Self:
        """
        Recursively transform the given data in to compatible types for JSONMapping.

        """
        if sub_mapping_cls is None:
            sub_mapping_cls_t: typing.Type[JSONMapping] = JSONMapping
        else:
            sub_mapping_cls_t = sub_mapping_cls

        @typing.overload
        def transform(value: typing.Mapping[str, typing.Any]) -> JSONMapping: ...

        @typing.overload
        def transform(value: typing.Any) -> FrozenJSONType: ...

        def transform(value: typing.Any) -> FrozenJSONType:
            if isinstance(value, Mapping) and not isinstance(value, JSONMapping):
                non_str_k = ", ".join(
                    sorted(
                        f"{repr(k)!r} (type {type(k).__name__})"
                        for k in value.keys()
                        if not isinstance(k, str)
                    ),
                )
                if non_str_k:
                    raise ValueError(
                        f"Unable to convert non-string key(s) to a valid frozen JSON "
                        f"type (got {non_str_k})",
                    )
                return sub_mapping_cls_t(
                    tuple((k, transform(v)) for k, v in value.items()),
                )
            if isinstance(value, (list, tuple)):
                return tuple(transform(v) for v in value)
            # For everything else, it must match what is allowed in FrozenJSONType.
            if isinstance(value, (str, int, float, bool, NoneType)):
                return value  # only_py37_type: ignore[return-value]

            raise ValueError(
                f"Unable to convert type {type(value).__name__} to a valid frozen JSON type",
            )

        return cls(transform(data))

    def __getitem__(self, key: str) -> FrozenJSONType:
        return self._data[key]

    def __iter__(self) -> typing.Iterator[str]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({repr(tuple(self._data.items()))})"

    def __or__(self, other: typing.Mapping[str, typing.Any]) -> Self:
        if not isinstance(other, type(self)):
            other = type(self).from_any_mapping(other)
        return type(self)({**self._data, **other})

    def __hash__(self) -> int:
        return hash(tuple(self.items()))



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import annotations

import asyncio
import typing

import packaging.version

from . import priority_selected
from .. import errors, model
from .._typing_compat import override


class MergedRepository(priority_selected.PrioritySelectedProjectsRepository):
    """
    Represents a merged view of all the given (unsorted) repositories

    NOTICE: The MergedRepository is combining the given repositories without
            giving exclusivity of a source of a specific package to any particular
            repositories. As a result, this implementation is vulnerable to
            dependency confusion. There are cases where this behaviour is desirable
            hence its existence, but if you are unsure of those reasons, consider
            using the :class:`PrioritySelectedProjectsRepository` instead.
    """

    @override
    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.ProjectDetail:
        """Retrieves a project page for the specified normalized project name
        by searching through the grouped list of sources and blending them together.
        """
        # Keep track of unique filenames for the merged files.
        files: typing.Dict[str, model.File] = {}

        results: typing.List[
            typing.Union[
                BaseException,
                model.ProjectDetail,
            ]
        ] = await asyncio.gather(
            *(
                source.get_project_page(
                    project_name,
                    request_context=request_context,
                )
                for source in self.sources
            ),
            return_exceptions=True,
        )

        project_pages: typing.List[model.ProjectDetail] = []
        for result in results:
            if isinstance(result, BaseException):
                if not isinstance(result, errors.PackageNotFoundError):
                    raise result
            else:
                for file in result.files:
                    # Only add the file if the filename hasn't been seen before.
                    files.setdefault(file.filename, file)
                project_pages.append(result)

        if not project_pages:
            raise errors.PackageNotFoundError(
                package_name=project_name,
            )

        # If we only have one resulting project page, bring it through verbatim,
        # and avoid the cost of having to compute the common metadata.
        if len(project_pages) == 1:
            return project_pages[0]

        # Downgrade the API version to the lowest available, as it will not be
        # possible to calculate the missing files to perform a version upgrade.
        api_version = min(
            packaging.version.Version(result.meta.api_version)
            for result in project_pages
        )

        versions = None
        if api_version >= packaging.version.Version("1.1"):
            # All project pages are >=1.1, therefore they MUST all have
            # versions specified. We combine them together.
            all_versions: typing.List[typing.FrozenSet[str]] = typing.cast(
                typing.List[typing.FrozenSet[str]],
                [page.versions for page in project_pages],
            )
            versions = frozenset().union(*all_versions)

        return model.ProjectDetail(
            meta=model.Meta(str(api_version)),
            name=project_pages[0].name,
            files=tuple(files.values()),
            versions=versions,
        )

    @override
    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.Resource:
        """
        Retrieves a resource from any source that has it.
        """
        project_found = False
        for source in self.sources:
            try:
                return await source.get_resource(
                    project_name,
                    resource_name,
                    request_context=request_context,
                )
            except errors.PackageNotFoundError:
                # Try the next source
                continue
            except errors.ResourceUnavailable:
                project_found = True
                # Try the next source
                continue

        # No source has the resource
        if not project_found:
            raise errors.PackageNotFoundError(project_name)
        else:
            raise errors.ResourceUnavailable(resource_name)



import pathlib
import shutil

import tomlkit

here = pathlib.Path(__file__).parent

build = here / 'built-repo'
if build.is_dir():
    shutil.rmtree(build)
build.mkdir()

srb = build / 'simple_repository'
shutil.copytree(here / 'simple_repository', srb)

for pcache in build.glob('**/__pycache__'):
    shutil.rmtree(pcache)

(srb / '_version.py').unlink()

for file in [here / 'pyproject.toml']:
    shutil.copy(file, build)

pyproject_path = build / 'pyproject.toml'
pyproject_content = tomlkit.parse(pyproject_path.read_text())
del pyproject_content['project']['maintainers']
pyproject_content['project']['urls']['Homepage'] = (
    "https://github.com/simple-repository/simple-repository"
)
pyproject_path.write_text(tomlkit.dumps(pyproject_content) + '\n')

for path in build.glob('**/*.py'):
    content = path.read_text()
    content = '\n'.join([
        '# Copyright (C) 2023, CERN',
        '# This software is distributed under the terms of the MIT',
        '# licence, copied verbatim in the file "LICENSE".',
        '# In applying this license, CERN does not waive the privileges and immunities',
        '# granted to it by virtue of its status as Intergovernmental Organization',
        '# or submit itself to any jurisdiction.',
        '',
    ]) + '\n' + content.lstrip()
    content = content.rstrip() + '\n'
    path.write_text(content, 'utf-8')



import pytest

from simple_repository._private_metadata import JSONMapping, PrivateMetadataMapping


def test__empty() -> None:
    assert PrivateMetadataMapping() == {}


def test__valid_names() -> None:
    mapping = PrivateMetadataMapping.from_any_mapping({"_a": 1, "_b": {"not_private": 1}})
    assert list(mapping.keys()) == ["_a", "_b"]
    assert type(mapping["_b"]) is JSONMapping


def test__invalid_names() -> None:
    with pytest.raises(ValueError, match="invalid for private metadata: 'a', 'b'"):
        PrivateMetadataMapping({"_f": 1, "a": 1, "b": True})



import re
import textwrap

import pytest

from simple_repository import _frozen_json


@pytest.fixture
def simple_json() -> _frozen_json.JSONMapping:
    return _frozen_json.JSONMapping(
        (
            ("string", "hello world 😀"),
            ("int", 1),
            ("bool", True),
            ("float", 3.14),
            ("null", None),
            ("tuple", (1, True, 3.14, None)),
            (
                "mapping",
                _frozen_json.JSONMapping(
                    (("int", 1), ("bool", True), ("float", 3.14), ("null", None)),
                ),
            ),
        ),
    )


def test_frozen_json_mapping__str(simple_json: _frozen_json.JSONMapping):
    assert simple_json["string"] == "hello world 😀"
    assert isinstance(simple_json["string"], str)


def test_frozen_json_mapping__int(simple_json: _frozen_json.JSONMapping):
    assert simple_json["int"] == 1
    assert isinstance(simple_json["int"], int)


def test_frozen_json_mapping__bool(simple_json: _frozen_json.JSONMapping):
    assert simple_json["bool"] is True


def test_frozen_json_mapping__float(simple_json: _frozen_json.JSONMapping):
    assert simple_json["float"] == 3.14
    assert isinstance(simple_json["float"], float)


def test_frozen_json_mapping__null(simple_json: _frozen_json.JSONMapping):
    assert simple_json["null"] is None


def test_frozen_json_mapping__tuple(simple_json: _frozen_json.JSONMapping):
    assert isinstance(simple_json["tuple"], tuple)


def test_frozen_json_mapping__mapping(simple_json: _frozen_json.JSONMapping):
    assert isinstance(simple_json["mapping"], _frozen_json.JSONMapping)
    assert list(simple_json["mapping"].keys()) == ["int", "bool", "float", "null"]


def test_frozen_json_mapping__init__with_none():
    m = _frozen_json.JSONMapping()
    assert tuple(m.items()) == ()


def test_frozen_json_mapping__from_any_mapping__with__list():
    m = _frozen_json.JSONMapping.from_any_mapping({"a": [1, 2, 3]})
    assert isinstance(m["a"], tuple)
    assert m["a"] == (1, 2, 3)


def test_frozen_json_mapping__from_any_mapping__with_dict():
    m = _frozen_json.JSONMapping.from_any_mapping({"a": {"b": [4, 5, None]}})
    assert isinstance(m["a"], _frozen_json.JSONMapping)
    assert isinstance(m["a"]["b"], tuple)
    assert m["a"]["b"] == (4, 5, None)


def test_frozen_json_mapping__init__from_any_mapping__with__dict_inside_tuple():
    m = _frozen_json.JSONMapping.from_any_mapping({"a": (1, 2, {})})
    assert isinstance(m["a"][2], _frozen_json.JSONMapping)


def test_frozen_json_mapping__init__from_any_mapping__with__circular_dict():
    data = {}
    data["data"] = data
    with pytest.raises(RecursionError):
        # It is not possible to have an immutable type that is recursive.
        _frozen_json.JSONMapping.from_any_mapping({"a": data})


def test_frozen_json_mapping__from_any_mapping__with_invalid_keys():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Unable to convert non-string key(s) to a valid frozen "
            "JSON type (got '1' (type int))",
        ),
    ):
        _frozen_json.JSONMapping.from_any_mapping({1: 1})  # type: ignore[arg-type]  # intentionally invalid for testing.


def test_frozen_json_mapping__from_any_mapping__with_invalid_type():
    class InvalidType:
        pass

    with pytest.raises(
        ValueError,
        match="Unable to convert type InvalidType to a valid frozen JSON type",
    ):
        _frozen_json.JSONMapping.from_any_mapping({"a": InvalidType()})


def test_frozen_json_mapping____str__(simple_json: _frozen_json.JSONMapping):
    assert (
        str(simple_json)
        == textwrap.dedent("""
    JSONMapping((('string', 'hello world 😀'), ('int', 1), ('bool', True), ('float', 3.14), ('null', None), ('tuple', (1, True, 3.14, None)), ('mapping', JSONMapping((('int', 1), ('bool', True), ('float', 3.14), ('null', None))))))
    """).strip()
    )


def test_frozen_json_mapping____repr__(simple_json: _frozen_json.JSONMapping):
    assert (
        repr(simple_json)
        == textwrap.dedent("""
    JSONMapping((('string', 'hello world 😀'), ('int', 1), ('bool', True), ('float', 3.14), ('null', None), ('tuple', (1, True, 3.14, None)), ('mapping', JSONMapping((('int', 1), ('bool', True), ('float', 3.14), ('null', None))))))
    """).strip()
    )


def test_frozen_json_mapping____hash__(simple_json: _frozen_json.JSONMapping):
    assert isinstance(hash(simple_json), int)


def test_frozen_json_mapping____hash___consistency():
    # Validate that the same object, produced at different times, produces the same hash.
    a = hash(_frozen_json.JSONMapping.from_any_mapping({"a": [1, 2, 3, {}]}))
    b = hash(_frozen_json.JSONMapping.from_any_mapping({"a": [1, 2, 3, {}]}))
    assert a == b


def test_frozen_json_mapping__or_operator(simple_json: _frozen_json.JSONMapping):
    new = simple_json | {"int": 2}
    assert new["int"] == 2
    assert simple_json["int"] == 1


def test_frozen_json_mapping__or_operator__ensure_frozen_too(simple_json: _frozen_json.JSONMapping):
    new = simple_json | {"some": {'mutable': ['things']}}
    assert isinstance(new["some"], _frozen_json.JSONMapping)
    assert isinstance(new["some"]['mutable'], tuple)
    assert new['some'] == {'mutable': ('things',)}


def test_frozen_json_mapping__drop(simple_json: _frozen_json.JSONMapping):
    new = _frozen_json.JSONMapping(
        {key: value for key, value in simple_json.items() if key not in ['int']},
    )
    assert simple_json["int"] == 1
    print('N:', new)
    with pytest.raises(KeyError):
        _ = new["int"]


def test_frozen_json_mapping__eq(simple_json: _frozen_json.JSONMapping):
    # __eq__ is given to us by the Mapping protocol, but quickly validate that it behaves as we expect
    # (https://docs.python.org/3/library/collections.abc.html#collections-abstract-base-classes)
    obj1 = _frozen_json.JSONMapping((("a", 1),))
    obj2 = _frozen_json.JSONMapping((("a", 1),))
    obj3 = _frozen_json.JSONMapping((("a", 2),))
    obj4 = _frozen_json.JSONMapping((("a", 1), ("b", 2)))

    assert obj1 == obj2
    assert obj1 != obj3
    assert obj1 != obj4



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import annotations

import pytest

from ... import model
from .fake_repository import FakeRepository


@pytest.fixture
def source_repository() -> FakeRepository:
    return FakeRepository(
        project_list=model.ProjectList(
            meta=model.Meta("1.0"),
            projects=frozenset([
                model.ProjectListElement("project1"),
                model.ProjectListElement("project2"),
                model.ProjectListElement("project3"),
            ]),
        ),
        project_pages=[
            model.ProjectDetail(
                meta=model.Meta("1.0"),
                name="project1",
                files=(),
            ),
            model.ProjectDetail(
                meta=model.Meta("1.0"),
                name="project2",
                files=(),
            ),
            model.ProjectDetail(
                meta=model.Meta("1.0"),
                name="project3",
                files=(),
            ),
        ],
        resources={
            "project1-1.0.tar.gz": model.HttpResource("content1"),
            "project2-1.0.tar.gz": model.HttpResource("content2"),
            "project3-1.0.tar.gz": model.HttpResource("content3"),
        },
    )



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import annotations

import typing

from .. import model
from .._typing_compat import override

if typing.TYPE_CHECKING:
    from .._typing_compat import TypeAlias
    WrappedFunction: TypeAlias = typing.Callable[..., typing.Any]


class SimpleRepository:
    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.ProjectDetail:
        """
        Get the project detail page of the given project

        The resulting project detail includes resources available for the
        project. These resources MUST be requested through the
        :meth:`get_resource` method, as the `model.File` resources may not be
        available through a URL.

        Parameters
        ----------
        project_name:
            The name of the project, which is not normalized.
        request_context:
            Additional meta information about the request being made. An example
            of such information is providing cache/ETag metadata to allow the
            repository to raise :class:`model.NotModified`.

        Returns
        -------
        model.ProjectDetail:
            The project detail content of the requested project

        Raises
        ------
        errors.PackageNotFoundError:
            If the project is not available in the repository.
        model.NotModified:
            When sufficient request context is provided, it is possible for the
            repository to raise a NotModified exception to indicate that the
            result that corresponds to the given cache headers is still valid.
        errors.UnsupportedSerialization:
            When the upstream repository is providing information that is not
            processable.
        errors.SourceRepositoryUnavailable:
            When the upstream repository is not available.
        """
        raise NotImplementedError()

    async def get_project_list(
        self,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.ProjectList:
        """
        Get the list of projects available in the repository.

        Parameters
        ----------
        request_context:
            Additional meta information about the request being made. An example
            of such information is providing cache/ETag metadata to allow the
            repository to raise :class:`model.NotModified`.

        Returns
        -------
        A list of all projects available on the repository.

        Raises
        ------
        model.NotModified:
            When sufficient request context is provided, it is possible for the
            repository to raise a NotModified exception to indicate that the
            result that corresponds to the given cache headers is still valid.
        errors.UnsupportedSerialization:
            When the upstream repository is providing information that is not
            processable.
        errors.SourceRepositoryUnavailable:
            When the upstream repository is not available.

        Notes
        -----
        It is technically possible for projects to be accessible
        from :meth:`get_project_page` which are not in the list provided by this
        method.

        """
        raise NotImplementedError()

    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.Resource:
        """
        Fetch a project detail resource from the repository

        Whilst this method is not part of the PEP-503 specification directly,
        this method exists such that resource requests can be processed by a
        repository component. Examples of useful processing that can happen
        include the computation of metadata upon request, and the ability to
        log specific requests.

        Parameters
        ----------
        project_name:
            The name of the project for which the resource is being requested.
        resource_name:
            The filename of the resource, as given by the ProjectDetail content.
        request_context:
            Additional meta information about the request being made. An example
            of such information is providing cache/ETag metadata to allow the
            repository to raise :class:`model.NotModified`.

        Raises
        ------
        errors.PackageNotFoundError:
            When the project does not exist in this repository.
        errors.ResourceUnavailable:
            When the project exists but the specific resource does not exist.
        model.NotModified:
            When sufficient request context is provided, it is possible for the
            repository to raise a NotModified exception to indicate that the
            result that corresponds to the given cache headers is still valid.
        errors.SourceRepositoryUnavailable:
            When the upstream repository is not available.

        Notes
        -----

        Repositories should distinguish between two failure cases:
        1. Project does not exist: raise PackageNotFoundError
        2. Project exists but resource does not exist: raise ResourceUnavailable

        """
        raise NotImplementedError()


class RepositoryContainer(SimpleRepository):
    """A base class for components that enhance the functionality of a source
    `SimpleRepository`. If not overridden, the methods provided by this class
    will delegate to the corresponding methods of the source repository.
    """
    def __init__(self, source: SimpleRepository) -> None:
        self.source = source

    @override
    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.ProjectDetail:
        return await self.source.get_project_page(project_name, request_context=request_context)

    @override
    async def get_project_list(
        self,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.ProjectList:
        return await self.source.get_project_list(request_context=request_context)

    @override
    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.Resource:
        return await self.source.get_resource(
            project_name,
            resource_name,
            request_context=request_context,
        )



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import annotations

import typing

from ... import errors, model
from ...components import core


class FakeRepository(core.SimpleRepository):

    def __init__(
        self,
        project_list: model.ProjectList = model.ProjectList(model.Meta('1.0'), frozenset()),
        project_pages: typing.Optional[typing.List[model.ProjectDetail]] = None,
        resources: typing.Optional[typing.Dict[str, model.Resource]] = None,
    ) -> None:
        self.project_list = project_list
        if project_pages:
            self.project_pages = {
                project.name: project for project in project_pages
            }
        else:
            self.project_pages = {}
        self.resources = resources or {}

    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.ProjectDetail:
        if project_name in self.project_pages:
            return self.project_pages[project_name]
        raise errors.PackageNotFoundError(project_name)

    async def get_project_list(
        self,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.ProjectList:
        return self.project_list

    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.Resource:
        # Check if we have this project first
        try:
            await self.get_project_page(project_name, request_context=request_context)
        except errors.PackageNotFoundError:
            raise

        # Look for the specific resource
        resource = self.resources.get(resource_name)
        if resource:
            return resource
        raise errors.ResourceUnavailable(resource_name)



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import annotations

import pytest

from ... import errors, model
from ...components.core import SimpleRepository
from ...components.merged import MergedRepository
from .fake_repository import FakeRepository


@pytest.mark.asyncio
async def test_get_project_page() -> None:
    repo = MergedRepository(
        [
            FakeRepository(),
            FakeRepository(
                project_pages=[
                    model.ProjectDetail(
                        model.Meta('1.1'),
                        "numpy",
                        files=(
                            model.File("numpy-1.1.whl", "url1", {}, size=1),
                            model.File("numpy-1.2.whl", "url1", {}, size=1),
                        ),
                    ),
                ],
            ),
            FakeRepository(
                project_pages=[
                    model.ProjectDetail(
                        model.Meta('1.0'),
                        "numpy",
                        files=(
                            model.File("numpy-1.1.whl", "url2", {}),
                            model.File("numpy-1.3.whl", "url2", {}),
                        ),
                    ),
                ],
            ),
        ],
    )

    resp = await repo.get_project_page("numpy")

    assert resp == model.ProjectDetail(
        model.Meta('1.0'),
        "numpy",
        files=(
            model.File("numpy-1.1.whl", "url1", {}, size=1),
            model.File("numpy-1.2.whl", "url1", {}, size=1),
            model.File("numpy-1.3.whl", "url2", {}),
        ),
    )


@pytest.mark.asyncio
async def test_get_project_page__versions() -> None:
    repo = MergedRepository(
        [
            FakeRepository(),
            FakeRepository(
                project_pages=[
                    model.ProjectDetail(
                        model.Meta('1.1'),
                        "numpy",
                        files=(
                            model.File("numpy-1.1.whl", "url1", {}, size=1),
                            model.File("numpy-1.2.whl", "url1", {}, size=1),
                        ),
                        versions=frozenset({"1.1", "1.2"}),
                    ),
                ],
            ),
            FakeRepository(
                project_pages=[
                    model.ProjectDetail(
                        model.Meta('1.2'),
                        "numpy",
                        files=(
                            model.File("numpy-1.1.whl", "url2", {}, size=1),
                            model.File("numpy-1.3.whl", "url2", {}, size=1),
                        ),
                        versions=frozenset({"1.1", "1.3", "1.5"}),
                    ),
                ],
            ),
        ],
    )

    resp = await repo.get_project_page("numpy")

    assert resp == model.ProjectDetail(
        model.Meta('1.1'),
        "numpy",
        files=(
            model.File("numpy-1.1.whl", "url1", {}, size=1),
            model.File("numpy-1.2.whl", "url1", {}, size=1),
            model.File("numpy-1.3.whl", "url2", {}, size=1),
        ),
        versions=frozenset({"1.1", "1.2", "1.3", "1.5"}),
    )


@pytest.mark.asyncio
async def test_get_project_page_failed() -> None:
    repo = MergedRepository([
        FakeRepository(),
        FakeRepository(),
        FakeRepository(),
    ])

    with pytest.raises(
        errors.PackageNotFoundError,
        match="Package 'numpy' was not found in the configured source",
    ):
        await repo.get_project_page("numpy")


@pytest.fixture
def resource_repo_t1() -> SimpleRepository:
    repo = MergedRepository([
        FakeRepository(
            project_pages=[
                model.ProjectDetail(
                    model.Meta("1.0"),
                    "numpy",
                    files=(),
                ),
            ],
            # First repo has project page but no resources
        ),
        FakeRepository(
            project_pages=[
                model.ProjectDetail(
                    model.Meta("1.0"),
                    "numpy",
                    files=(),
                ),
            ],
            resources={
                "numpy-1.0.whl": model.HttpResource("from_second_repo"),
            },
        ),
        FakeRepository(
            project_pages=[
                model.ProjectDetail(
                    model.Meta("1.0"),
                    "numpy",
                    files=(),
                ),
            ],
            resources={
                "numpy-1.0.whl": model.HttpResource("from_third_repo"),
                "numpy-1.1.whl": model.HttpResource("from_third_repo"),
                "numpy-1.0.whl.metadata": model.HttpResource("from_third_repo"),
            },
        ),
    ])
    return repo


@pytest.mark.asyncio
async def test_get_resource__wrong_project_name(resource_repo_t1: SimpleRepository) -> None:
    # Should find resource from second repo (first one that has it).
    resource = await resource_repo_t1.get_resource("numpy", "numpy-1.0.whl")
    assert resource == model.HttpResource("from_second_repo")


@pytest.mark.asyncio
async def test_get_resource__searches_all_sources(resource_repo_t1: SimpleRepository) -> None:
    # Should find resource from second repo even though first repo exists
    resource = await resource_repo_t1.get_resource("numpy", "numpy-1.0.whl")
    assert resource == model.HttpResource("from_second_repo")

    resource = await resource_repo_t1.get_resource("numpy", "numpy-1.0.whl.metadata")
    assert resource == model.HttpResource("from_third_repo")


@pytest.mark.asyncio
async def test_get_resource__not_found(resource_repo_t1: SimpleRepository) -> None:
    with pytest.raises(errors.ResourceUnavailable, match="missing.whl"):
        await resource_repo_t1.get_resource("numpy", "missing.whl")

    with pytest.raises(errors.PackageNotFoundError, match="missing-project"):
        await resource_repo_t1.get_resource("missing-project", "missing.whl")



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import annotations

import dataclasses

import pytest

from .. import model


def test_ProjectListElement__normalized_name() -> None:
    prj = model.ProjectListElement('some-.name')
    assert prj.normalized_name == 'some-name'


def test_ProjectDetail__normalized_name() -> None:
    project_detail = model.ProjectDetail(
        meta=model.Meta("1.0"),
        name="Project",
        files=(),
    )
    assert project_detail._normalized_name == "project"


def test_ProjectDetail__failed_post_init() -> None:
    with pytest.raises(
        ValueError,
        match="SimpleAPI>=1.1 requires the size field to be set for all the files.",
    ):
        model.ProjectDetail(
            meta=model.Meta("1.1"),
            name="pippo",
            files=(
                model.File(
                    filename="pippozzo",
                    url="url",
                    hashes={},
                ),
            ),
        )


def test_ProjectDetail__post_init_v1() -> None:
    project_detail = model.ProjectDetail(
        meta=model.Meta("1.0"),
        name="pippo",
        files=(
            model.File(
                filename="pippozzo",
                url="url",
                hashes={},
            ),
        ),
    )
    assert project_detail.versions is None


def test_ProjectDetail__post_init_v1_1() -> None:
    project_detail = model.ProjectDetail(
        meta=model.Meta("1.1"),
        name="pippo",
        files=(
            model.File(
                filename="pippo-1.0.tar.gz",
                url="url",
                hashes={},
                size=1,
            ),
            model.File(
                filename="pippo-2.0-anylinux-py3.whl",
                url="url",
                hashes={},
                size=1,
            ),
        ),
    )
    assert project_detail.versions == {
        "1.0", "2.0",
    }


def test_ProjectDetail__versions_subset() -> None:
    detail = model.ProjectDetail(
        meta=model.Meta("1.1"),
        name="pippo",
        files=(
            model.File(
                filename="pippo-1.0.tar.gz",
                url="url",
                hashes={},
                size=1,
            ),
            model.File(
                filename="pippo-2.0-anylinux-py3.whl",
                url="url",
                hashes={},
                size=1,
            ),
        ),
        versions=frozenset({'1.0'}),
    )
    # When we have a subset of versions specified (not spec compliant), we don't
    # do any magic, but we also don't raise.
    assert detail.versions == {'1.0'}


def test_ProjectDetail__manual_versions() -> None:
    # Sometimes we want to be able to state that there are versions with no
    # files (PEP-700):
    # > The versions key MAY contain versions with no associated files
    project_detail = model.ProjectDetail(
        meta=model.Meta("1.1"),
        name="pippo",
        files=(
            model.File(
                filename="pippo-1.0.tar.gz",
                url="url",
                hashes={},
                size=1,
            ),
            model.File(
                filename="pippo-2.0-anylinux-py3.whl",
                url="url",
                hashes={},
                size=1,
            ),
        ),
        versions=frozenset({'1.0', '2.0', "1.2.3"}),
    )
    # Versions should persist with other replacement.
    pd1 = dataclasses.replace(project_detail, files=project_detail.files[:1])

    assert pd1.versions == {'1.0', '2.0', "1.2.3"}

    # And we should be able to get it to be generated again by setting it to None.
    pd2 = dataclasses.replace(project_detail, versions=None)
    assert pd2.versions == {"1.0", "2.0"}


def test__File__arbitrary_private_metadata() -> None:
    file = model.File(
        filename="pippo",
        url="url",
        hashes={},
        private_metadata=model.PrivateMetadataMapping(dict(_foo='bar')),
    )
    assert dict(file.private_metadata) == {'_foo': 'bar'}
    # Ensure that the private attributes survives additional public attribute
    # changes.
    new_file = dataclasses.replace(file, filename='bar')
    assert dict(new_file.private_metadata) == {'_foo': 'bar'}


def test__File__eq__private_metadata() -> None:
    file = model.File(
        filename="pippo",
        url="url",
        hashes={},
        private_metadata=model.PrivateMetadataMapping(dict(_foo='bar')),
    )
    file2 = model.File(
        filename="pippo",
        url="url",
        hashes={},
    )
    assert file != file2
    assert file == file


@pytest.mark.xfail(strict=True)
def test__File__hash__private_metadata() -> None:
    file = model.File(
        filename="pippo",
        url="url",
        hashes={},
        private_metadata=model.PrivateMetadataMapping(dict(_foo='bar')),
    )
    file2 = model.File(
        filename="pippo",
        url="url",
        hashes={},
    )
    assert hash(file) != hash(file2)


@pytest.mark.xfail(strict=True)
def test__File__hash() -> None:
    file = model.File(
        filename="pippo",
        url="url",
        hashes={},
    )
    # The file should be hashable (it is frozen after all), but we currently
    # allow dict to be passed.
    hash(file)



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import annotations

import dataclasses
from datetime import datetime, timedelta
import typing

from . import core
from .. import model
from .._typing_compat import override


class NewReleasesRemover(core.RepositoryContainer):
    """
    A component used to remove newly released projects from the source
    repository until they have existed for the given quarantine time.
    This component can be used only if the source repository exposes the upload
    date according to PEP-700: https://peps.python.org/pep-0700/.
    """
    def __init__(
        self,
        source: core.SimpleRepository,
        quarantine_time: timedelta = timedelta(days=2),
        whitelist: typing.Tuple[str, ...] = (),
    ) -> None:
        self._quarantine_time = quarantine_time
        self._whitelist = whitelist
        super().__init__(source)

    @override
    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: typing.Optional[model.RequestContext] = None,
    ) -> model.ProjectDetail:
        project_page = await super().get_project_page(
            project_name,
            request_context=request_context,
        )

        if project_name in self._whitelist:
            return project_page

        return self._exclude_recent_distributions(
            project_page=project_page,
            now=datetime.now(),
        )

    def _exclude_recent_distributions(
        self,
        project_page: model.ProjectDetail,
        now: datetime,
    ) -> model.ProjectDetail:
        files_to_maintain = []
        files_to_be_removed: typing.List[typing.Tuple[model.File, datetime]] = []

        for file in project_page.files:
            if not file.upload_time:
                # We maintain the file if there is no upload time information.
                files_to_maintain.append(file)
            else:
                quarantine_release_time = file.upload_time + self._quarantine_time
                # Maintain the file if it has been available for longer than the quarantine time.
                if quarantine_release_time < now:
                    files_to_maintain.append(file)
                else:
                    files_to_be_removed.append((file, quarantine_release_time))

        date_format = "%Y-%m-%dT%H:%M:%SZ"
        serialized_quarantined_files = [
            {
                'filename': file.filename,
                'quarantine_release_time': quarantine_release_time.strftime(date_format),
                'upload_time': typing.cast(datetime, file.upload_time).strftime(date_format),
            } for file, quarantine_release_time in files_to_be_removed
        ]
        # Note that we don't remove the version from the versions list on project page.
        # PEP-700 states that we are allowed to have a release without any files in it.
        return dataclasses.replace(
            project_page,
            files=tuple(files_to_maintain),
            # Use a private attribute to give context of the files that have been quarantined.
            private_metadata=project_page.private_metadata | {
                '_quarantined_files': serialized_quarantined_files,
            },
        )



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import sys

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from mock import AsyncMock

__all__ = ['AsyncMock']



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import annotations

from datetime import datetime
import logging
import os
import pathlib
from unittest import mock

import pytest

from ... import errors, model
from ...components import core
from ...components.resource_cache import ResourceCacheRepository
from .fake_repository import FakeRepository
from .mock_compat import AsyncMock


@pytest.fixture
def repository(tmp_path: pathlib.Path) -> ResourceCacheRepository:
    http_resource = model.HttpResource("url/http-1.0-any.whl")
    http_resource.context["etag"] = "etag"

    not_to_cache_resource = model.HttpResource("url/uncacheable-1.0-any.whl", to_cache=False)
    not_to_cache_resource.context["etag"] = "etag"
    source = FakeRepository(
        project_pages=[
            model.ProjectDetail(model.Meta("1.0"), "http", files=()),
            model.ProjectDetail(model.Meta("1.0"), "local", files=()),
            model.ProjectDetail(model.Meta("1.0"), "http-no-etag", files=()),
            model.ProjectDetail(model.Meta("1.0"), "uncacheable", files=()),
            model.ProjectDetail(model.Meta("1.0"), "resource", files=()),
        ],
        resources={
            "http-1.0-any.whl": http_resource,
            "local-1.0.tar.gz": model.LocalResource(pathlib.Path("path")),
            "http_no_etag-1.0-any.whl": model.HttpResource("url/http_no_etag-1.0-any.whl"),
            "uncacheable-1.0-any.whl": not_to_cache_resource,
        },
    )
    return ResourceCacheRepository(
        source=source,
        cache_path=tmp_path,
        http_client=mock.MagicMock(),
    )


@pytest.mark.asyncio
async def test_get_resource__cache_hit(repository: ResourceCacheRepository) -> None:
    (repository._cache_path / "http").mkdir()
    cached_file = repository._cache_path / "http" / "http-1.0-any.whl"
    cached_file.write_text("cached content")
    cached_info_file = repository._cache_path / "http" / "http-1.0-any.whl.info"
    # The content of the the info file matches the upstream cache
    cached_info_file.write_text("etag")

    resource = await repository.get_resource(
        project_name="http",
        resource_name="http-1.0-any.whl",
    )
    # The cache returns a LocalResource pointing to
    # the cached file with the same etag as upstream
    assert isinstance(resource, model.LocalResource)
    assert resource.path == cached_file
    assert resource.context["etag"] == "etag"
    assert cached_file.read_text() == "cached content"


@pytest.mark.asyncio
async def test_get_resource__cache_miss__wrong_etag(repository: ResourceCacheRepository) -> None:
    (repository._cache_path / "http").mkdir()
    cached_file = repository._cache_path / "http" / "http-1.0-any.whl"
    cached_file.write_text("invalid cached content")
    cached_info_file = repository._cache_path / "http" / "http-1.0-any.whl.info"
    # The content of the the info file matches the upstream cache
    cached_info_file.write_text("wrong_etag")

    with mock.patch(
        "simple_repository.utils.download_file",
        AsyncMock(
            side_effect=lambda **kwargs: kwargs["dest_file"].touch(),
        ),
    ):
        response = await repository.get_resource(
            project_name="http",
            resource_name="http-1.0-any.whl",
        )

    # after the cache miss, the upstream file is downloaded and the info file is created.
    # The cache returns a LocalResource pointing to the downloaded file.
    assert isinstance(response, model.LocalResource)
    assert response.path == repository._cache_path / "http" / "http-1.0-any.whl"
    assert (repository._cache_path / "http" / "http-1.0-any.whl.info").read_text() == "etag"
    assert (repository._cache_path / "http" / "http-1.0-any.whl").read_text() != "invalid cached content"


@pytest.mark.asyncio
async def test_get_resource__cache_miss__no_etag(repository: ResourceCacheRepository) -> None:
    assert not (repository._cache_path / "http" / "http-1.0-any.whl.info").is_file()
    assert not (repository._cache_path / "http" / "http-1.0-any.whl").is_file()

    with mock.patch(
        "simple_repository.utils.download_file",
        AsyncMock(
            side_effect=lambda **kwargs: kwargs["dest_file"].touch(),
        ),
    ):
        response = await repository.get_resource(
            project_name="http",
            resource_name="http-1.0-any.whl",
        )

    # after the cache miss, the upstream file is downloaded and the info file is created.
    # The cache returns a LocalResource pointing to the downloaded file.
    assert isinstance(response, model.LocalResource)
    assert response.path == repository._cache_path / "http" / "http-1.0-any.whl"
    assert (repository._cache_path / "http" / "http-1.0-any.whl.info").is_file()
    assert (repository._cache_path / "http" / "http-1.0-any.whl").is_file()
    assert response.context.get("etag") == "etag"


@pytest.mark.asyncio
async def test_get_resource__cache_miss_local_resource(
    repository: ResourceCacheRepository,
) -> None:
    resource = await repository.get_resource(
        project_name="local",
        resource_name="local-1.0.tar.gz",
    )

    # the cache doesn't store LocalResources. No cache file or info file is created.
    assert isinstance(resource, model.LocalResource)
    assert resource.path == pathlib.Path("path")
    assert not (repository._cache_path / "local" / "local-1.0.tar.gz.info").is_file()
    assert not (repository._cache_path / "local" / "local-1.0.tar.gz").is_file()


@pytest.mark.asyncio
async def test_get_resource__path_traversal(
    repository: ResourceCacheRepository,
) -> None:
    with pytest.raises(
        ValueError,
        match="is not contained in",
    ):
        await repository.get_resource(
            project_name="not-used",
            resource_name="../../../etc/passwords",
        )


@pytest.mark.asyncio
async def test_get_resource__source_unavailable_cache_hit(
    tmp_path: pathlib.Path,
) -> None:
    source = AsyncMock(
        spec=core.SimpleRepository,
        get_resource=AsyncMock(side_effect=errors.SourceRepositoryUnavailable),
    )

    (tmp_path / "project-name").mkdir()

    (tmp_path / "project-name" / "resource-name.info").write_text("etag")
    (tmp_path / "project-name" / "resource-name").write_text("content")

    repository = ResourceCacheRepository(
        source=source,
        cache_path=tmp_path,
        http_client=mock.MagicMock(),
    )

    resource = await repository.get_resource("project-name", "resource-name")
    assert isinstance(resource, model.LocalResource)
    assert resource.path == tmp_path / "project-name" / "resource-name"


@pytest.mark.asyncio
async def test_get_resource__source_unavailable_cache_hit__falback_disabled(
    tmp_path: pathlib.Path,
) -> None:
    source = AsyncMock(
        spec=core.SimpleRepository,
        get_resource=AsyncMock(side_effect=errors.SourceRepositoryUnavailable),
    )

    (tmp_path / "project-name").mkdir()

    (tmp_path / "project-name" / "resource-name.info").write_text("etag")
    (tmp_path / "project-name" / "resource-name").write_text("content")

    repository = ResourceCacheRepository(
        source=source,
        cache_path=tmp_path,
        http_client=mock.MagicMock(),
        fallback_to_cache=False,
    )

    with pytest.raises(errors.SourceRepositoryUnavailable):
        await repository.get_resource("project-name", "resource-name")


@pytest.mark.asyncio
async def test_get_resource__source_unavailable_cache_hit__log(
    tmp_path: pathlib.Path,
) -> None:
    source = AsyncMock(
        spec=core.SimpleRepository,
        get_resource=AsyncMock(side_effect=errors.SourceRepositoryUnavailable),
    )

    (tmp_path / "project-name").mkdir()

    (tmp_path / "project-name" / "resource-name.info").write_text("etag")
    (tmp_path / "project-name" / "resource-name").write_text("content")

    mock_logger = mock.Mock(spec=logging.Logger)

    repository = ResourceCacheRepository(
        source=source,
        cache_path=tmp_path,
        http_client=mock.MagicMock(),
        logger=mock_logger,
    )

    await repository.get_resource("project-name", "resource-name")
    mock_logger.error.assert_called_once_with(
        "Upstream unavailable, served cached project-name:resource-name",
    )


@pytest.mark.asyncio
async def test_get_resource__source_unavailable_cache_miss(
    tmp_path: pathlib.Path,
) -> None:
    source = AsyncMock(
        spec=core.SimpleRepository,
        get_resource=AsyncMock(side_effect=errors.SourceRepositoryUnavailable),
    )

    repository = ResourceCacheRepository(
        source=source,
        cache_path=tmp_path,
        http_client=mock.MagicMock(),
    )

    with pytest.raises(errors.SourceRepositoryUnavailable):
        await repository.get_resource("project-name", "resource-name")


@pytest.mark.asyncio
async def test_get_resource__no_cache_created_when_no_upstream_etag_exists(
    repository: ResourceCacheRepository,
) -> None:
    resource = await repository.get_resource(
        project_name="http-no-etag",
        resource_name="http_no_etag-1.0-any.whl",
    )

    # Upstream doesn't set an etag so the HttpResource is not cached,
    # otherwise a LocalResource would have been returned
    # No cache file or info file gets created and an http resource is returned.
    assert isinstance(resource, model.HttpResource)
    assert not (repository._cache_path / "http-no-etag" / "http_no_etag-1.0-any.whl").is_file()
    assert not (repository._cache_path / "http-no-etag" / "http_no_etag-1.0-any.whl.info").is_file()


@pytest.mark.parametrize(
    "resource", [
        model.HttpResource(url="url", to_cache=False),
        model.LocalResource(path=pathlib.Path("."), to_cache=False),
        model.TextResource(text="text", to_cache=False),
    ],
)
@pytest.mark.asyncio
async def test_get_resource__no_cache_created_when_to_cache_is_false(
    resource: model.Resource,
    tmp_path: pathlib.Path,
) -> None:
    resource.context["etag"] = "etag"

    repository = ResourceCacheRepository(
        source=FakeRepository(
            project_pages=[
                model.ProjectDetail(model.Meta("1.0"), "resource", files=()),
            ],
            resources={"resource-1.0-any.whl": resource},
        ),
        cache_path=tmp_path,
        http_client=mock.MagicMock(),
    )

    result = await repository.get_resource(
        project_name="resource",
        resource_name="resource-1.0-any.whl",
    )

    # Upstream sets an etag but the to_cache attribute is False so the resource is not cached.
    # No cache file or info file gets created and an same resource is returned.
    assert resource == result
    assert not (repository._cache_path / "resource" / "resource-1.0-any.whl").is_file()
    assert not (repository._cache_path / "resource" / "resource-1.0-any.whl.info").is_file()


@pytest.mark.asyncio
async def test_get_resource__source_raised_not_modified__request_etag_invalid(
    repository: ResourceCacheRepository,
) -> None:
    # Upstream raised not modified and the request context misses the etag.
    repository.source = mock.Mock(get_resource=AsyncMock(side_effect=model.NotModified))

    (repository._cache_path / "http").mkdir()
    cached_file = repository._cache_path / "http" / "http-1.0-any.whl"
    cached_file.write_text("cached content")
    cached_info_file = repository._cache_path / "http" / "http-1.0-any.whl.info"
    # The content of the the info file matches the upstream cache
    cached_info_file.write_text("etag")

    resource = await repository.get_resource(
        project_name="http",
        resource_name="http-1.0-any.whl",
    )
    # The cache returns a LocalResource pointing to
    # the cached file with the same etag as upstream
    assert isinstance(resource, model.LocalResource)
    assert resource.path == cached_file
    assert resource.context["etag"] == "etag"
    assert cached_file.read_text() == "cached content"


@pytest.mark.asyncio
async def test_get_resource__source_raised_not_modified__request_etag_valid(
    repository: ResourceCacheRepository,
) -> None:
    # Upstream raised not modified and the request etag matches.
    repository.source = mock.Mock(get_resource=AsyncMock(side_effect=model.NotModified))

    (repository._cache_path / "http").mkdir()
    cached_file = repository._cache_path / "http" / "http-1.0-any.whl"
    cached_file.write_text("cached content")
    cached_info_file = repository._cache_path / "http" / "http-1.0-any.whl.info"
    # The content of the the info file matches the upstream cache
    cached_info_file.write_text("etag")

    context = model.RequestContext({"etag": "etag"})

    with pytest.raises(model.NotModified):
        await repository.get_resource(
            project_name="http",
            resource_name="http-1.0-any.whl",
            request_context=context,
        )


def test_resource_cache_init(tmp_path: pathlib.Path) -> None:
    real_repo = tmp_path / "dir" / "cache_repo"
    real_repo.mkdir(parents=True)

    symlink = tmp_path / "link"
    symlink.symlink_to(real_repo)

    repo = ResourceCacheRepository(
        source=AsyncMock(),
        cache_path=symlink,
        http_client=mock.MagicMock(),
    )
    assert str(symlink) != str(real_repo)
    assert str(repo._cache_path) == str(real_repo)


def test_update_last_access(repository: ResourceCacheRepository) -> None:
    cached_info = repository._cache_path / "my_resource.ext.info"
    cached_info.touch()

    with mock.patch(
        "simple_repository.components.resource_cache.datetime",
        mock.Mock(
            now=mock.Mock(return_value=datetime.fromisoformat("2006-07-09")),
            fromisoformat=datetime.fromisoformat,
            spec=datetime,
        ),
    ):
        repository._update_last_access(cached_info)

    assert os.path.getatime(cached_info) == datetime.fromisoformat("2006-07-09").timestamp()

    with mock.patch(
        "simple_repository.components.resource_cache.datetime",
        mock.Mock(
            now=mock.Mock(return_value=datetime.fromisoformat("2025-07-09")),
            fromisoformat=datetime.fromisoformat,
            spec=datetime,
        ),
    ):
        repository._update_last_access(cached_info)

    assert os.path.getatime(cached_info) == datetime.fromisoformat("2025-07-09").timestamp()


@pytest.mark.asyncio
async def test_update_last_access__cache_hit_called(
    repository: ResourceCacheRepository,
) -> None:
    (repository._cache_path / "http").mkdir()
    cached_file = (repository._cache_path / "http" / "http-1.0-any.whl")
    cached_file.touch()
    cached_info_file = repository._cache_path / "http" / "http-1.0-any.whl.info"
    cached_info_file.write_text("etag")

    update_last_access_mock = mock.Mock()

    with mock.patch.object(
        target=ResourceCacheRepository,
        attribute="_update_last_access",
        new=update_last_access_mock,
    ):
        await repository.get_resource(
            project_name="http",
            resource_name="http-1.0-any.whl",
        )

    update_last_access_mock.assert_called_once()


@pytest.mark.asyncio
async def test_update_last_access_for__cache_miss_local_not_called(
    repository: ResourceCacheRepository,
) -> None:
    update_last_access_for_mock = mock.Mock()
    with mock.patch.object(
        target=ResourceCacheRepository,
        attribute="_update_last_access",
        new=update_last_access_for_mock,
    ):
        await repository.get_resource(
            project_name="local",
            resource_name="local-1.0.tar.gz",
        )
    update_last_access_for_mock.assert_not_called()


@pytest.mark.asyncio
async def test_update_last_access_for__cache_miss_remote_called(
    repository: ResourceCacheRepository,
) -> None:
    update_last_access_for_mock = mock.Mock()
    with mock.patch(
        "simple_repository.utils.download_file",
        AsyncMock(
            side_effect=lambda **kwargs: kwargs["dest_file"].touch(),
        ),
    ), mock.patch.object(
        target=ResourceCacheRepository,
        attribute="_update_last_access",
        new=update_last_access_for_mock,
    ):
        await repository.get_resource(
            project_name="http",
            resource_name="http-1.0-any.whl",
        )
    update_last_access_for_mock.assert_called_once()


@pytest.mark.asyncio
async def test_store_resource__http_resource(
    repository: ResourceCacheRepository,
    tmp_path: pathlib.Path,
) -> None:
    http_resource = model.HttpResource(url="my_url")
    cache_path = tmp_path / "cache"

    with mock.patch(
        "simple_repository.utils.download_file",
        AsyncMock(
            side_effect=lambda **kwargs: kwargs["dest_file"].write_text("http content"),
        ),
    ):
        await repository._store_resource(
            resource=http_resource,
            upstream_etag="etag",
            resource_path=cache_path,
            resource_info_path=cache_path.with_suffix(".metadata"),
        )

    assert cache_path.read_text() == "http content"
    assert cache_path.with_suffix(".metadata").read_text() == "etag"


@pytest.mark.asyncio
async def test_store_resource__local_resource(
    repository: ResourceCacheRepository,
    tmp_path: pathlib.Path,
) -> None:
    file_path = tmp_path / "file"
    (tmp_path / "file").write_text("file_content")
    cache_path = tmp_path / "cache"
    local_resource = model.LocalResource(file_path)

    await repository._store_resource(
        resource=local_resource,
        upstream_etag="etag",
        resource_path=cache_path,
        resource_info_path=cache_path.with_suffix(".metadata"),
    )

    assert cache_path.read_text() == "file_content"
    assert cache_path.with_suffix(".metadata").read_text() == "etag"


@pytest.mark.asyncio
async def test_store_resource__text_resource(
    repository: ResourceCacheRepository,
    tmp_path: pathlib.Path,
) -> None:
    cache_path = tmp_path / "cache"
    text_resource = model.TextResource("my_text")

    await repository._store_resource(
        resource=text_resource,
        upstream_etag="etag",
        resource_path=cache_path,
        resource_info_path=cache_path.with_suffix(".metadata"),
    )

    assert cache_path.read_text() == "my_text"
    assert cache_path.with_suffix(".metadata").read_text() == "etag"



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import annotations

from datetime import datetime
import json
import typing

import pytest

from .. import model, utils
from ..serializer import SerializerHtmlV1, SerializerJsonV1


def test_serialize_file_html() -> None:
    serializer = SerializerHtmlV1()

    file = model.File(
        filename="test.html",
        url="https://example.com/test.html",
        hashes={"test": "123", "sha256": "abc123"},
        requires_python=">=3.6",
    )
    expected = (
        '<a href="https://example.com/test.html#sha256=abc123" data-requires-python="&gt;=3.6"'
        '>test.html</a><br/>\n'
    )
    assert serializer._serialize_file(file) == expected

    file = model.File(
        filename="test.html",
        url="https://example.com/test.html",
        hashes={},
    )
    expected = (
        '<a href="https://example.com/test.html"'
        '>test.html</a><br/>\n'
    )
    assert serializer._serialize_file(file) == expected


@pytest.mark.parametrize(
    "yank_attr, yank_value",
    [
        (' data-yanked=""', True),
        (' data-yanked="reason"', "reason"),
        ('', None),
        ('', False),
    ],
)
def test_serialize_file_html_yank(
    yank_attr: str,
    yank_value: typing.Union[bool, str, None],
) -> None:
    serializer = SerializerHtmlV1(

    )
    file = model.File(
        filename="test.html",
        url="https://example.com/test.html",
        hashes={},
        yanked=yank_value,
    )
    expected = (
        '<a href="https://example.com/test.html"'
        f'{yank_attr}'
        '>test.html</a><br/>\n'
    )
    assert serializer._serialize_file(file) == expected


@pytest.mark.parametrize(
    "metadata_attr, metadata_value",
    [
        (' data-core-metadata="true"', True),
        (' data-core-metadata="sha=..."', {"sha": "..."}),
        ('', None),
        ('', False),
    ],
)
def test_serialize_file_html_metadata(
    metadata_attr: str,
    metadata_value: typing.Union[bool, typing.Dict[str, str], None],
) -> None:
    serializer = SerializerHtmlV1()

    file = model.File(
        filename="test.html",
        url="https://example.com/test.html",
        hashes={},
        dist_info_metadata=metadata_value,
    )
    expected = (
        '<a href="https://example.com/test.html"'
        f'{metadata_attr}'
        '>test.html</a><br/>\n'
    )
    assert serializer._serialize_file(file) == expected


@pytest.mark.parametrize(
    "gpg_attr, gpg_value",
    [
        (' data-gpg-sig="true"', True),
        (' data-gpg-sig="false"', False),
        ('', None),
    ],
)
def test_serialize_file_html_gpg(gpg_attr: str, gpg_value: typing.Optional[bool]) -> None:
    serializer = SerializerHtmlV1()

    file = model.File(
        filename="test.html",
        url="https://example.com/test.html",
        hashes={},
        gpg_sig=gpg_value,
    )
    expected = (
        '<a href="https://example.com/test.html"'
        f'{gpg_attr}'
        '>test.html</a><br/>\n'
    )
    assert serializer._serialize_file(file) == expected


def test_serialize_project_page_html() -> None:
    project_page = model.ProjectDetail(
        meta=model.Meta(api_version="1.0"),
        name="test-project",
        files=(
            model.File(filename="test.html", url="https://example.com/test.html", hashes={}),
            model.File(filename="test.txt", url="test.txt", hashes={}),
        ),
    )
    expected = """<!DOCTYPE html>
    <html>
    <head>
        <meta name="pypi:repository-version" content="1.0">
        <title>Links for test-project</title>
    </head>
    <body>
    <h1>Links for test-project</h1>
<a href="https://example.com/test.html">test.html</a><br/>
<a href="test.txt">test.txt</a><br/>
</body>
</html>"""
    serializer = SerializerHtmlV1()
    assert serializer.serialize_project_page(project_page) == expected


def test_serialize_project_list_html() -> None:
    project_list = model.ProjectList(
        meta=model.Meta(api_version="1.0"),
        projects=frozenset([
            model.ProjectListElement(name="test-project-1"),
            model.ProjectListElement(name="test-project-2"),
        ]),
    )
    expected_header = """<!DOCTYPE html>
    <html>
    <head>
        <meta name="pypi:repository-version" content="1.0">
        <title>Simple index</title>
    </head>
    <body>
"""
    expected_footer = """</body>
</html>"""

    serializer = SerializerHtmlV1()
    serialized_page = serializer.serialize_project_list(project_list)
    assert serialized_page.startswith(expected_header)
    assert serialized_page.endswith(expected_footer)
    a_tags = utils.remove_suffix(
        utils.remove_prefix(serialized_page, expected_header),
        expected_footer,
    )
    assert '<a href="test-project-1/">test-project-1</a><br/>' in a_tags
    assert '<a href="test-project-2/">test-project-2</a><br/>' in a_tags


def test_serialize_project_page_json() -> None:
    page = model.ProjectDetail(
        model.Meta("1.0"),
        "project",
        files=(
            model.File(
                filename="test1.whl",
                url="test1.whl",
                hashes={},
            ),
            model.File(
                filename="test2.whl",
                url="test2.whl",
                hashes={"hash": "test_hash"},
                requires_python=">4.0",
                dist_info_metadata=True,
                yanked="yanked",
                gpg_sig=True,
            ),
            model.File(
                filename="test3.whl",
                url="test3.whl",
                hashes={},
                dist_info_metadata={"sha": "..."},
                yanked=True,
            ),
        ),
    )
    serializer = SerializerJsonV1()
    res = serializer.serialize_project_page(page)

    assert json.loads(res) == json.loads('''
        {
            "meta": {
                "api-version": "1.0"
            },
            "name": "project",
            "files": [
                {
                    "filename": "test1.whl",
                    "url": "test1.whl",
                    "hashes": {}
                },
                {
                    "filename": "test2.whl",
                    "url": "test2.whl",
                    "hashes": {"hash": "test_hash"},
                    "requires-python": ">4.0",
                    "core-metadata": true,
                    "yanked": "yanked",
                    "gpg-sig": true
                },
                {
                    "filename": "test3.whl",
                    "url": "test3.whl",
                    "hashes": {},
                    "core-metadata": {"sha": "..."},
                    "yanked": true
                }
            ]
        }
    ''')


@pytest.mark.parametrize(
    "version, serialization", [
        (
            "1.0",
            '''[{
                "filename": "test1-1.0.whl",
                "url": "test1.whl",
                "hashes": {}
            }]''',
        ),
        (
            "1.1",
            '''[{
                "filename": "test1-1.0.whl",
                "url": "test1.whl",
                "hashes": {},
                "size": 1,
                "upload-time": "2000-01-04T00:00:00Z"
            }]''',
        ),
    ],
)
def test_serialize_project_page_json__v1_1_attrs(
    version: str,
    serialization: str,
) -> None:
    page = model.ProjectDetail(
        model.Meta(version),
        "project",
        files=(
            model.File(
                filename="test1-1.0.whl",
                url="test1.whl",
                hashes={},
                size=1,
                upload_time=datetime(2000, 1, 4, 0, 0, 0),
            ),
        ),
    )
    serializer = SerializerJsonV1()
    res = serializer.serialize_project_page(page)
    assert json.loads(res)["files"] == json.loads(serialization)


def test_serialize_project_page_json__private_attrs() -> None:
    serialization = '''{
      "meta": {
        "api-version": "1.1",
        "_meta_extra": "abc"
      },
      "name": "project",
      "versions": [
        "1.2.3"
      ],
      "files": [
        {
          "filename": "test1-1.2.3-any.whl",
          "url": "test1.whl",
          "hashes": {},
          "size": 1,
          "_file_extra": 123
        }
      ],
      "_page_extra": 456
    }'''
    page = model.ProjectDetail(
        model.Meta("1.1", private_metadata=model.PrivateMetadataMapping(dict(_meta_extra="abc"))),
        "project",
        files=(
            model.File(
                filename="test1-1.2.3-any.whl",
                url="test1.whl",
                hashes={},
                size=1,
                private_metadata=model.PrivateMetadataMapping(dict(_file_extra=123)),
            ),
        ),
        private_metadata=model.PrivateMetadataMapping(dict(_page_extra=456)),
    )
    serializer = SerializerJsonV1()
    res = serializer.serialize_project_page(page)
    assert json.loads(res) == json.loads(serialization)


def test_serialize_project_list_json() -> None:
    page = model.ProjectList(
        model.Meta("1.0"),
        projects=frozenset([
            model.ProjectListElement("a"),
        ]),
    )
    serializer = SerializerJsonV1()
    res = serializer.serialize_project_list(page)

    assert res == json.dumps(
        json.loads('''{
        "meta": {
            "api-version": "1.0"
        },
        "projects": [
            {"name": "a"}
        ]
    }'''),
    )


def test_serialize_project_list_json__with_extra() -> None:
    page = model.ProjectList(
        model.Meta("1.0", private_metadata=model.PrivateMetadataMapping(dict(_extra_meta="abc"))),
        projects=frozenset([
            model.ProjectListElement(
                "a",
                private_metadata=model.PrivateMetadataMapping.from_any_mapping(
                    dict(_extra_list_element=123, _nested={"nested": True}),
                ),
            ),
        ]),
        private_metadata=model.PrivateMetadataMapping(dict(_extra_project_list=456)),
    )
    serializer = SerializerJsonV1()
    res = serializer.serialize_project_list(page)

    assert res == json.dumps(
        json.loads('''{
        "meta": {
            "api-version": "1.0",
            "_extra_meta": "abc"
        },
        "projects": [
            {"name": "a", "_extra_list_element": 123, "_nested": {"nested": true}}
        ],
        "_extra_project_list": 456
    }'''),
    )
